@@ -17,64 +17,75 @@
 # under the License.
 
 import collections
-import decorator
-import random
-import time
+import logging
 
-import itertools
+from concurrent import futures
 import msgpack
 import pymemcache.client
+import retrying
 import six
 
 from tooz import coordination
 from tooz import locking
+from tooz import utils
+
+
+LOG = logging.getLogger(__name__)
 
 
 class Retry(Exception):
     """Exception raised if we need to retry."""
 
 
-@decorator.decorator
-def retry(f, *args, **kwargs):
-    for try_number in itertools.chain(six.moves.range(10),
-                                      itertools.repeat(10)):
-        try:
-            return f(*args, **kwargs)
-        except Retry:
-            pass
-        time.sleep(random.randint(0, (2 ** try_number)) / 1000.0)
+def retry_if_retry_raised(exception):
+    return isinstance(exception, Retry)
+
+
+def retry(f):
+    return retrying.retry(
+        retry_on_exception=retry_if_retry_raised,
+        wait='exponential_sleep', wait_exponential_max=1)(f)
 
 
 class MemcachedLock(locking.Lock):
     _LOCK_PREFIX = b'__TOOZ_LOCK_'
 
     def __init__(self, coord, name, timeout):
+        super(MemcachedLock, self).__init__(self._LOCK_PREFIX + name)
         self.coord = coord
-        self.name = self._LOCK_PREFIX + name
         self.timeout = timeout
 
+    @retry
     def acquire(self, blocking=True):
-        while True:
-            if self.coord.client.add(
-                    self.name,
-                    self.coord._member_id,
-                    expire=self.timeout,
-                    noreply=False):
-                self.coord._acquired_locks.append(self)
-                return True
-            if not blocking:
-                return False
-            # NOTE(jd) Configurable? :/
-            time.sleep(0.11)
+        if self.coord.client.add(
+                self.name,
+                self.coord._member_id,
+                expire=self.timeout,
+                noreply=False):
+            self.coord._acquired_locks.append(self)
+            return True
+        if not blocking:
+            return False
+        raise Retry
 
     def release(self):
-        self.coord._acquired_locks.remove(self)
-        self.coord.client.delete(self.name)
+        if self.coord.client.delete(self.name, noreply=False):
+            self.coord._acquired_locks.remove(self)
+            return True
+        else:
+            return False
 
     def heartbeat(self):
         """Keep the lock alive."""
-        self.coord.client.touch(self.name,
-                                expire=self.timeout)
+        poked = self.coord.client.touch(self.name,
+                                        expire=self.timeout,
+                                        noreply=False)
+        if not poked:
+            LOG.warn("Unable to heartbeat by updating key '%s' with extended"
+                     " expiry of %s seconds", self.name, self.timeout)
+
+    def get_owner(self):
+        return self.coord.client.get(self.name)
 
 
 class MemcachedDriver(coordination.CoordinationDriver):
@@ -84,14 +95,21 @@ class MemcachedDriver(coordination.CoordinationDriver):
     _MEMBER_PREFIX = b'_TOOZ_MEMBER_'
     _GROUP_LIST_KEY = b'_TOOZ_GROUP_LIST'
 
-    def __init__(self, member_id, membership_timeout=30, lock_timeout=30,
-                 leader_timeout=30):
+    def __init__(self, member_id, parsed_url, options):
         super(MemcachedDriver, self).__init__()
         self._member_id = member_id
         self._groups = set()
-        self.membership_timeout = membership_timeout
-        self.lock_timeout = lock_timeout
-        self.leader_timeout = leader_timeout
+        self._executor = None
+        self.host = (parsed_url.hostname or "localhost",
+                     parsed_url.port or 11211)
+        default_timeout = options.get('timeout', ['30'])
+        self.timeout = int(default_timeout[-1])
+        self.membership_timeout = int(options.get(
+            'membership_timeout', default_timeout)[-1])
+        self.lock_timeout = int(options.get(
+            'lock_timeout', default_timeout)[-1])
+        self.leader_timeout = int(options.get(
+            'leader_timeout', default_timeout)[-1])
 
     @staticmethod
     def _msgpack_serializer(key, value):
@@ -107,28 +125,32 @@ class MemcachedDriver(coordination.CoordinationDriver):
             return msgpack.loads(value)
         raise Exception("Unknown serialization format")
 
-    def start(self, host=("127.0.0.1", 11211), timeout=5):
+    def start(self):
         try:
             self.client = pymemcache.client.Client(
-                host,
+                self.host,
                 serializer=self._msgpack_serializer,
                 deserializer=self._msgpack_deserializer,
-                timeout=timeout,
-                connect_timeout=timeout)
+                timeout=self.timeout,
+                connect_timeout=self.timeout)
         except Exception as e:
-            raise coordination.ToozConnectionError(e)
+            raise coordination.ToozConnectionError(utils.exception_message(e))
         self._group_members = collections.defaultdict(set)
         self._acquired_locks = []
+        self._executor = futures.ThreadPoolExecutor(max_workers=1)
         self.heartbeat()
 
     def stop(self):
+        if self._executor is not None:
+            self._executor.shutdown(wait=True)
+            self._executor = None
+
+        for lock in list(self._acquired_locks):
+            lock.release()
+
         self.client.delete(self._encode_member_id(self._member_id))
         map(self.leave_group, list(self._groups))
 
-        for group_id in six.iterkeys(self._hooks_elected_leader):
-            if self.get_leader(group_id).get() == self._member_id:
-                self.client.delete(self._encode_group_leader(group_id))
-
         self.client.close()
 
     def _encode_group_id(self, group_id):
@@ -162,15 +184,18 @@ class MemcachedDriver(coordination.CoordinationDriver):
 
     def create_group(self, group_id):
         encoded_group = self._encode_group_id(group_id)
-        if not self.client.add(encoded_group, {}, noreply=False):
-            return MemcachedAsyncError(
-                coordination.GroupAlreadyExist(group_id))
-        self._add_group_to_group_list(group_id)
-        return MemcachedAsyncResult(None)
+
+        def _create_group():
+            if not self.client.add(encoded_group, {}, noreply=False):
+                raise coordination.GroupAlreadyExist(group_id)
+            self._add_group_to_group_list(group_id)
+
+        return MemcachedFutureResult(self._executor.submit(_create_group))
 
     def get_groups(self):
-        return MemcachedAsyncResult(
-            self.client.get(self._GROUP_LIST_KEY) or [])
+        def _get_groups():
+            return self.client.get(self._GROUP_LIST_KEY) or []
+        return MemcachedFutureResult(self._executor.submit(_get_groups))
 
     @retry
     def join_group(self, group_id, capabilities=b""):
@@ -221,20 +246,18 @@ class MemcachedDriver(coordination.CoordinationDriver):
                     if self.client.get(self._encode_member_id(m)))
 
     def get_members(self, group_id):
-        try:
-            return MemcachedAsyncResult(self._get_members(group_id).keys())
-        except Exception as e:
-            return MemcachedAsyncError(e)
+        def _get_members():
+            return self._get_members(group_id).keys()
+        return MemcachedFutureResult(self._executor.submit(_get_members))
 
     def get_member_capabilities(self, group_id, member_id):
-        try:
+        def _get_member_capabilities():
             group_members = self._get_members(group_id)
-        except Exception as e:
-            return MemcachedAsyncError(e)
-        if member_id not in group_members:
-            return MemcachedAsyncError(
-                coordination.MemberNotJoined(group_id, member_id))
-        return MemcachedAsyncResult(group_members[member_id][b'capabilities'])
+            if member_id not in group_members:
+                raise coordination.MemberNotJoined(group_id, member_id)
+            return group_members[member_id][b'capabilities']
+        return MemcachedFutureResult(
+            self._executor.submit(_get_member_capabilities))
 
     @retry
     def update_capabilities(self, group_id, capabilities):
@@ -253,8 +276,9 @@ class MemcachedDriver(coordination.CoordinationDriver):
         return MemcachedAsyncResult(None)
 
     def get_leader(self, group_id):
-        return MemcachedAsyncResult(
-            self.client.get(self._encode_group_leader(group_id)))
+        def _get_leader():
+            return self._get_leader_lock(group_id).get_owner()
+        return MemcachedFutureResult(self._executor.submit(_get_leader))
 
     def heartbeat(self):
         self.client.set(self._encode_member_id(self._member_id),
@@ -264,11 +288,6 @@ class MemcachedDriver(coordination.CoordinationDriver):
         for lock in self._acquired_locks:
             lock.heartbeat()
 
-        for group_id in six.iterkeys(self._hooks_elected_leader):
-            if self.get_leader(group_id).get() == self._member_id:
-                self.client.touch(self._encode_group_leader(group_id),
-                                  expire=self.leader_timeout)
-
     def _init_watch_group(self, group_id):
         members = self.client.get(self._encode_group_id(group_id))
         if members is None:
@@ -306,6 +325,10 @@ class MemcachedDriver(coordination.CoordinationDriver):
     def get_lock(self, name):
         return MemcachedLock(self, name, self.lock_timeout)
 
+    def _get_leader_lock(self, group_id):
+        return MemcachedLock(self, self._encode_group_leader(group_id),
+                             self.leader_timeout)
+
     def run_watchers(self):
         result = []
         for group_id in self.client.get(self._GROUP_LIST_KEY):
@@ -327,22 +350,30 @@ class MemcachedDriver(coordination.CoordinationDriver):
 
             self._group_members[group_id] = group_members
 
-        for group_id in six.iterkeys(self._hooks_elected_leader):
-            lock_id = self._encode_group_leader(group_id)
+        for group_id, hooks in six.iteritems(self._hooks_elected_leader):
             # Try to grab the lock, if that fails, that means someone has it
             # already.
-            if self.client.add(lock_id, self._member_id,
-                               expire=self.leader_timeout,
-                               noreply=False):
+            if self._get_leader_lock(group_id).acquire(blocking=False):
                 # We got the lock
-                self._hooks_elected_leader[group_id].run(
-                    coordination.LeaderElected(
-                        group_id,
-                        self._member_id))
+                hooks.run(coordination.LeaderElected(
+                    group_id,
+                    self._member_id))
 
         return result
 
 
+class MemcachedFutureResult(coordination.CoordAsyncResult):
+    """Memcached asynchronous result that references a future."""
+    def __init__(self, fut):
+        self._fut = fut
+
+    def get(self, timeout=10):
+        return self._fut.result(timeout=timeout)
+
+    def done(self):
+        return self._fut.done()
+
+
 class MemcachedAsyncResult(coordination.CoordAsyncResult):
     """Memcached asynchronous result.
 
@@ -351,10 +382,10 @@ class MemcachedAsyncResult(coordination.CoordAsyncResult):
 
     """
     def __init__(self, result):
-        self.result = result
+        self._result = result
 
     def get(self, timeout=0):
-        return self.result
+        return self._result
 
     @staticmethod
     def done():
@@ -369,10 +400,10 @@ class MemcachedAsyncError(coordination.CoordAsyncResult):
 
     """
     def __init__(self, error):
-        self.error = error
+        self._error = error
 
-    def get(self, timeout=0):
-        raise self.error
+    def get(self, timeout=10):
+        raise self._error
 
     @staticmethod
     def done():
